import numpy as np
from gym import utils
from gym.envs.mujoco import mujoco_env
from numpy.linalg import norm

DEFAULT_CAMERA_CONFIG = {
    'distance': 4.0,
}
scale = 35.

class MultiGoalAntEnv(mujoco_env.MujocoEnv, utils.EzPickle):
    def __init__(self,
                 xml_file='ant.xml',
                 ctrl_cost_weight=0.5,
                 contact_cost_weight=5e-4,
                 healthy_reward=0.0,
                 terminate_when_unhealthy=True,
                 healthy_z_range=(0.2, 1.0),
                 contact_force_range=(-1.0, 1.0),
                 reset_noise_scale=0.1,
                 exclude_current_positions_from_observation=False):
        utils.EzPickle.__init__(**locals())

        self.threshold = 1.5
        self.gs =  np.array([
            [30,0], [0,30],
            [-30,0], [0,-30]], np.float32)
        self.r_coeff = 100.
        self.g_coeff = 10.

        self.p = np.array([0, 0], np.float32)
        self.idx = 0
        self.potential_prev = self.potential

        self._ctrl_cost_weight = ctrl_cost_weight
        self._contact_cost_weight = contact_cost_weight

        self._healthy_reward = healthy_reward
        self._terminate_when_unhealthy = terminate_when_unhealthy
        self._healthy_z_range = healthy_z_range

        self._contact_force_range = contact_force_range

        self._reset_noise_scale = reset_noise_scale

        self._exclude_current_positions_from_observation = (
            exclude_current_positions_from_observation)

        mujoco_env.MujocoEnv.__init__(self, xml_file, 5)

    @property
    def healthy_reward(self):
        return float(
            self.is_healthy
            or self._terminate_when_unhealthy
        ) * self._healthy_reward

    def control_cost(self, action):
        control_cost = self._ctrl_cost_weight * np.sum(np.square(action))
        return control_cost

    @property
    def contact_forces(self):
        raw_contact_forces = self.sim.data.cfrc_ext
        min_value, max_value = self._contact_force_range
        contact_forces = np.clip(raw_contact_forces, min_value, max_value)
        return contact_forces

    @property
    def contact_cost(self):
        contact_cost = self._contact_cost_weight * np.sum(
            np.square(self.contact_forces))
        return contact_cost

    @property
    def is_healthy(self):
        state = self.state_vector()
        min_z, max_z = self._healthy_z_range
        is_healthy = (np.isfinite(state).all() and min_z <= state[2] <= max_z)
        return is_healthy

    @property
    def done(self):
        done = self.is_goal or (not self.is_healthy
                if self._terminate_when_unhealthy
                else False)
        return done

    def step(self, action):
        xy_position_before = self.get_body_com("torso")[:2].copy()
        self.do_simulation(action, self.frame_skip)
        xy_position_after = self.get_body_com("torso")[:2].copy()
        d_before = norm(xy_position_before - self.gs[self.idx])
        d_after = norm(xy_position_after - self.gs[self.idx])
        #g_velocity = (d_after - d_before) / self.dt
        g_velocity = (d_before - d_after) / self.dt

        xy_velocity = (xy_position_after - xy_position_before) / self.dt
        x_velocity, y_velocity = xy_velocity

        ctrl_cost = self.control_cost(action)
        contact_cost = self.contact_cost


        self.p = xy_position_after

        # pot = self.potential
        # potential_diff = pot - self.potential_prev
        # self.potential_prev = pot
        # distance_r = self.r_coeff * potential_diff
        self.renew_idx()


        goal_bonus = self.g_coeff if self.is_goal else 0.

        forward_reward = g_velocity

        healthy_reward = self.healthy_reward

        rewards = forward_reward# + healthy_reward + goal_bonus
        costs = ctrl_cost + contact_cost

        reward = rewards - costs
        done = self.done
        observation = self._get_obs()
        info = {
            'reward_forward': forward_reward,
            'reward_ctrl': -ctrl_cost,
            'reward_contact': -contact_cost,
            'reward_survive': healthy_reward,

            'x_position': xy_position_after[0],
            'y_position': xy_position_after[1],
            'distance_from_origin': norm(xy_position_after, ord=2),

            'x_velocity': x_velocity,
            'y_velocity': y_velocity,
            'forward_reward': forward_reward,
        }

        return observation, reward, done, info

    def _get_obs(self):
        position = self.sim.data.qpos.flat.copy()
        velocity = self.sim.data.qvel.flat.copy()
        contact_force = self.contact_forces.flat.copy()

        if self._exclude_current_positions_from_observation:
            position = position[2:]

        observations = np.concatenate((position, velocity, contact_force))

        return observations


    @property
    def distance(self):
        idx = self.idx
        g = self.gs[idx]
        d = norm(self.p - g)
        return d

    @property
    def is_goal(self):
        return self.distance <= self.threshold

    def renew_idx(self):
        self.idx = np.argmin(norm(self.p[None] - self.gs, axis=1))
        return self.idx

    @property
    def potential(self):
        ds = norm(self.p[None] - self.gs, axis=1)
        gau = np.exp(-.5 * np.square(ds/scale))
        return np.sum(gau)


    @property
    def distance(self):
        idx = self.idx
        g = self.gs[idx]
        d = norm(self.p - g)
        return d

    def reset_model(self):
        noise_low = -self._reset_noise_scale
        noise_high = self._reset_noise_scale

        qpos = self.init_qpos + self.np_random.uniform(
            low=noise_low, high=noise_high, size=self.model.nq)
        qvel = self.init_qvel + self._reset_noise_scale * self.np_random.randn(
            self.model.nv)
        self.set_state(qpos, qvel)

        observation = self._get_obs()

        self.p = self.get_body_com("torso")[:2].copy()
        self.renew_idx()
        self.potential_prev = self.potential

        return observation

    def viewer_setup(self):
        for key, value in DEFAULT_CAMERA_CONFIG.items():
            if isinstance(value, np.ndarray):
                getattr(self.viewer.cam, key)[:] = value
            else:
                setattr(self.viewer.cam, key, value)
